from __future__ import annotations

import asyncio
import itertools
import logging
import re
import time
import traceback
import uuid
from collections import defaultdict
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

import phonenumbers
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from packaging.version import Version

from signalbot.api import ReceiveMessagesError, SignalAPI
from signalbot.command import Command
from signalbot.context import Context
from signalbot.message import Message, UnknownMessageFormatError
from signalbot.storage import RedisStorage, SQLiteStorage

if TYPE_CHECKING:
    from signalbot.link_previews import LinkPreview

CommandList: TypeAlias = list[
    tuple[
        Command,
        list[str] | bool,
        list[str] | bool,
        Callable[[Message], bool] | None,
    ]
]

LOGGER_NAME = "signalbot"


def enable_console_logging(level: int = logging.WARNING) -> None:
    handler = logging.StreamHandler()

    formatter = logging.Formatter(
        "%(asctime)s %(name)s [%(levelname)s] - %(funcName)s - %(message)s"
    )
    handler.setFormatter(formatter)

    logger = logging.getLogger(LOGGER_NAME)
    logger.addHandler(handler)
    logger.setLevel(level)


class SignalBot:
    def __init__(self, config: dict):  # noqa: ANN204
        """SignalBot

        Example Config:
        ======= Mandatory fields ========
        signal_service: "127.0.0.1:8080"
        phone_number: "+49123456789"

        ======= Optional fields ========
        storage:
            redis_host: "redis"
            redis_port: 6379
        retry_interval: 1
        download_attachments: True
        """
        self._logger = logging.getLogger(LOGGER_NAME)

        self.config = config

        self._commands_to_be_registered: CommandList = []  # populated by .register()
        self.commands: CommandList = []  # populated by .start()

        self.groups = []  # populated by .start()
        self._groups_by_id = {}
        self._groups_by_internal_id = {}
        self._groups_by_name = defaultdict(list)

        try:
            self._phone_number = self.config["phone_number"]
            self._signal_service = self.config["signal_service"]
            download_attachments = self.config.get("download_attachments", True)
            self._signal = SignalAPI(
                self._signal_service,
                self._phone_number,
                download_attachments,
            )
        except KeyError:
            raise SignalBotError("Could not initialize SignalAPI with given config")  # noqa: B904, EM101, TRY003

        self._event_loop = asyncio.get_event_loop()
        self._q = asyncio.Queue()
        self._running_tasks: set[asyncio.Task] = set()

        self._produce_tasks: set[asyncio.Task] = set()
        self._consume_tasks: set[asyncio.Task] = set()

        try:
            self.scheduler = AsyncIOScheduler(event_loop=self._event_loop)
        except Exception as e:  # noqa: BLE001
            raise SignalBotError(f"Could not initialize scheduler: {e}")  # noqa: B904, EM102, TRY003

        config_storage = {}
        try:
            config_storage = self.config["storage"]
            if config_storage.get("type") == "sqlite":
                self._sqlite_db = config_storage["sqlite_db"]
                check_same_thread = config_storage.get("check_same_thread", True)
                self.storage = SQLiteStorage(
                    self._sqlite_db,
                    check_same_thread=check_same_thread,
                )
                self._logger.info("sqlite storage initilized")
            else:
                self._redis_host = config_storage["redis_host"]
                self._redis_port = config_storage["redis_port"]
                self.storage = RedisStorage(self._redis_host, self._redis_port)
                self._logger.info("redis storage initilized")
        except Exception:  # noqa: BLE001
            self.storage = SQLiteStorage()
            if config_storage.get("type") != "in-memory":
                self._logger.warning(
                    "[Bot] Could not initialize Redis and no SQLite DB name was given."
                    " In-memory storage will be used."
                    " Restarting will delete the storage!"
                    " Add storage: {'type': 'in-memory'}"
                    " to the config to silence this error.",
                )
            if "redis_host" in config_storage:
                self._logger.warning(
                    f"[Bot] Redis initialization error: {traceback.format_exc()}",  # noqa: G004
                )

    def register(
        self,
        command: Command,
        contacts: list[str] | bool = True,  # noqa: FBT001, FBT002
        groups: list[str] | bool = True,  # noqa: FBT001, FBT002
        f: Callable[[Message], bool] | None = None,
    ) -> None:
        command.bot = self
        command.setup()
        self._commands_to_be_registered.append((command, contacts, groups, f))

    async def _resolve_commands(self) -> None:
        self.commands = []
        for command, contacts, groups, f in self._commands_to_be_registered:
            group_ids = None

            if isinstance(groups, bool):
                group_ids = groups

            if isinstance(groups, list):
                group_ids = []
                for group in groups:
                    if self._is_group_id(group):  # group is a group id, higher prio
                        group_ids.append(group)
                    else:  # group is a group name
                        matched_group = self._get_group_by_name(group)
                        if matched_group is not None:
                            group_ids.append(matched_group["id"])
                        else:
                            self._logger.warning(
                                f"[Bot] [{command.__class__.__name__}] '{group}' is not a valid group name or id",  # noqa: E501, G004
                            )

            self.commands.append((command, contacts, group_ids, f))

    async def _async_post_init(self) -> None:
        await self._check_signal_service()
        await self._check_signal_cli_rest_api_version()
        await self._detect_groups()
        await self._resolve_commands()
        await self._produce_consume_messages()

    async def _check_signal_service(self) -> None:
        while (await self._signal.check_signal_service()) is False:
            self._logger.error(
                "Cannot connect to the signal-cli-rest-api service, retrying"
            )
            await asyncio.sleep(self.config.get("retry_interval", 1))

    async def _check_signal_cli_rest_api_version(self) -> None:
        min_version = Version("0.95.0")
        version = await self._signal.get_signal_cli_rest_api_version()
        if Version(version) < min_version:
            raise RuntimeError(  # noqa: TRY003
                f"Incompatible signal-cli-rest-api version, found {version}, minimum required is {min_version}",  # noqa: E501, EM102
            )

    def _store_reference_to_task(
        self,
        task: asyncio.Task,
        task_set: set[asyncio.Task],
    ) -> None:
        # Keep a hard reference to the tasks, fixes Ruff's RUF006 rule
        task_set.add(task)
        task.add_done_callback(task_set.discard)

    def start(self, run_forever: bool = True) -> None:  # noqa: FBT001, FBT002
        task = self._event_loop.create_task(
            self._rerun_on_exception(self._async_post_init),
        )
        self._store_reference_to_task(task, self._running_tasks)

        if run_forever:
            # Add more scheduler tasks here
            # self.scheduler.add_job(...)
            self.scheduler.start()

            self._event_loop.run_forever()

    async def send(  # noqa: PLR0913
        self,
        receiver: str,
        text: str,
        *,
        base64_attachments: list | None = None,
        link_preview: LinkPreview | None = None,
        quote_author: str | None = None,
        quote_mentions: list | None = None,
        quote_message: str | None = None,
        quote_timestamp: int | None = None,
        mentions: (
            list[dict[str, Any]] | None
        ) = None,  # [{ "author": "uuid" , "start": 0, "length": 1 }]
        edit_timestamp: int | None = None,
        text_mode: str | None = None,
        view_once: bool = False,
    ) -> int:
        receiver = self._resolve_receiver(receiver)
        link_preview_raw = link_preview.model_dump() if link_preview else None

        resp = await self._signal.send(
            receiver,
            text,
            base64_attachments=base64_attachments,
            link_preview=link_preview_raw,
            quote_author=quote_author,
            quote_mentions=quote_mentions,
            quote_message=quote_message,
            quote_timestamp=quote_timestamp,
            mentions=mentions,
            text_mode=text_mode,
            edit_timestamp=edit_timestamp,
            view_once=view_once,
        )
        resp_payload = await resp.json()
        timestamp = int(resp_payload["timestamp"])
        self._logger.info(f"[Bot] New message {timestamp} sent:\n{text}")  # noqa: G004

        return timestamp

    async def react(self, message: Message, emoji: str) -> None:
        # TODO: check that emoji is really an emoji  # noqa: TD002, TD003
        recipient = message.recipient()
        recipient = self._resolve_receiver(recipient)
        target_author = message.source
        timestamp = message.timestamp
        await self._signal.react(recipient, emoji, target_author, timestamp)
        self._logger.info(f"[Bot] New reaction: {emoji}")  # noqa: G004

    async def receipt(
        self,
        message: Message,
        receipt_type: Literal["read", "viewed"],
    ) -> None:
        if message.group is not None:
            self._logger.warning("[Bot] Receipts are not supported for groups")
            return

        recipient = self._resolve_receiver(message.recipient())
        await self._signal.receipt(recipient, receipt_type, message.timestamp)
        self._logger.info(f"[Bot] Receipt: {receipt_type}")  # noqa: G004

    async def start_typing(self, receiver: str) -> None:
        receiver = self._resolve_receiver(receiver)
        await self._signal.start_typing(receiver)

    async def stop_typing(self, receiver: str) -> None:
        receiver = self._resolve_receiver(receiver)
        await self._signal.stop_typing(receiver)

    async def update_contact(
        self,
        receiver: str,
        expiration_in_seconds: int | None = None,
        name: str | None = None,
    ) -> None:
        receiver = self._resolve_receiver(receiver)
        await self._signal.update_contact(
            receiver,
            expiration_in_seconds=expiration_in_seconds,
            name=name,
        )

    async def update_group(
        self,
        group_id: str,
        base64_avatar: str | None = None,
        description: str | None = None,
        expiration_in_seconds: int | None = None,
        name: str | None = None,
    ) -> None:
        group_id = self._resolve_receiver(group_id)
        await self._signal.update_group(
            group_id,
            base64_avatar=base64_avatar,
            description=description,
            expiration_in_seconds=expiration_in_seconds,
            name=name,
        )

    async def remote_delete(self, receiver: str, timestamp: int) -> int:
        receiver = self._resolve_receiver(receiver)

        resp = await self._signal.remote_delete(
            receiver,
            timestamp=timestamp,
        )
        resp_payload = await resp.json()
        ret_timestamp = int(resp_payload["timestamp"])
        self._logger.info(f"[Bot] Deleted message with timestamp {timestamp}")  # noqa: G004

        return ret_timestamp

    async def delete_attachment(self, attachment_filename: str) -> None:
        # Delete the attachment from the local storage
        await self._signal.delete_attachment(attachment_filename)

    async def _detect_groups(self) -> None:
        # reset group lookups to avoid stale data
        self.groups = await self._signal.get_groups()

        self._groups_by_id: dict[str, dict[str, Any]] = {}
        self._groups_by_internal_id: dict[str, dict[str, Any]] = {}
        self._groups_by_name: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)
        for group in self.groups:
            self._groups_by_id[group["id"]] = group
            self._groups_by_internal_id[group["internal_id"]] = group
            self._groups_by_name[group["name"]].append(group)

        self._logger.info(f"[Bot] {len(self.groups)} groups detected")  # noqa: G004

    def _resolve_receiver(self, receiver: str) -> str:
        if self._is_phone_number(receiver):
            return receiver

        if self._is_valid_uuid(receiver):
            return receiver

        if self._is_username(receiver):
            return receiver

        if self._is_group_id(receiver):
            return receiver

        group = self._groups_by_internal_id.get(receiver)
        if group is not None:
            return group["id"]

        group = self._get_group_by_name(receiver)
        if group is not None:
            return group["id"]

        raise SignalBotError("Cannot resolve receiver.")  # noqa: EM101, TRY003

    def _is_phone_number(self, phone_number: str) -> bool:
        try:
            parsed_number = phonenumbers.parse(phone_number, region=None)
            return phonenumbers.is_valid_number(parsed_number)
        except phonenumbers.phonenumberutil.NumberParseException:
            return False

    def _is_valid_uuid(self, receiver_uuid: str) -> bool:
        try:
            uuid.UUID(str(receiver_uuid))
            return True  # noqa: TRY300
        except ValueError:
            return False

    def _is_username(self, receiver_username: str) -> bool:  # noqa: PLR0911
        """
        Check if username has correct format, as described in
        https://support.signal.org/hc/en-us/articles/6712070553754-Phone-Number-Privacy-and-Usernames#username_req
        Additionally, cannot have more than 9 digits and the digits cannot be 00.
        """
        split_username = receiver_username.split(".")
        if len(split_username) == 2:  # noqa: PLR2004
            characters = split_username[0]
            digits = split_username[1]
            if len(characters) < 3 or len(characters) > 32:  # noqa: PLR2004
                return False
            if not re.match(r"^[A-Za-z\d_]+$", characters):
                return False
            if len(digits) < 2 or len(digits) > 9:  # noqa: PLR2004
                return False
            try:
                digits = int(digits)
                if digits == 0:  # noqa: SIM103
                    return False
                return True  # noqa: TRY300
            except ValueError:
                return False
        else:
            return False

    def _is_group_id(self, group_id: str) -> bool:
        """Check if group_id has the right format, e.g.

              random string                                              length 66
              ↓                                                          ↓
        group.OyZzqio1xDmYiLsQ1VsqRcUFOU4tK2TcECmYt2KeozHJwglMBHAPS7jlkrm=
        ↑                                                                ↑
        prefix                                                           suffix
        """
        if group_id is None:
            return False

        return re.match(r"^group\.[a-zA-Z0-9]{59}=$", group_id)

    def _is_internal_id(self, internal_id: str) -> bool:
        if internal_id is None:
            return False
        return internal_id[-1] == "="

    def _get_group_by_name(self, group_name: str) -> dict[str, Any] | None:
        groups = self._groups_by_name.get(group_name)
        if groups is not None:
            if len(groups) > 1:
                self._logger.warning(
                    f"[Bot] There is more than one group named '{group_name}', using the first one.",  # noqa: E501, G004
                )
            return groups[0]
        return None

    # see https://stackoverflow.com/questions/55184226/catching-exceptions-in-individual-tasks-and-restarting-them
    async def _rerun_on_exception(self, coro, *args, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN202
        """Restart coroutine by waiting an exponential time deplay"""
        max_sleep = 5 * 60  # sleep for at most 5 mins until rerun
        reset = 3 * 60  # reset after 3 minutes running successfully
        init_sleep = 1  # always start with sleeping for 1 second

        next_sleep = init_sleep
        while True:
            start_t = int(time.monotonic())  # seconds

            try:
                return await coro(*args, **kwargs)
            except asyncio.CancelledError:
                raise
            except Exception:  # noqa: BLE001
                traceback.print_exc()

            end_t = int(time.monotonic())  # seconds

            if end_t - start_t < reset:
                sleep_t = next_sleep
                next_sleep = min(max_sleep, next_sleep * 2)  # double sleep time
            else:
                next_sleep = init_sleep  # reset sleep time
                sleep_t = next_sleep

            self._logger.warning(f"Restarting coroutine in {sleep_t} seconds")  # noqa: G004
            await asyncio.sleep(sleep_t)

    async def _produce_consume_messages(
        self,
        producers: int = 1,
        consumers: int = 3,
    ) -> None:
        for task in itertools.chain(self._consume_tasks, self._produce_tasks):
            task.cancel()

        self._produce_tasks.clear()

        for n in range(1, producers + 1):
            produce_task = self._rerun_on_exception(self._produce, n)
            produce_task = asyncio.create_task(produce_task)
            self._store_reference_to_task(produce_task, self._produce_tasks)

        self._consume_tasks.clear()

        for n in range(1, consumers + 1):
            consume_task = self._rerun_on_exception(self._consume, n)
            consume_task = asyncio.create_task(consume_task)
            self._store_reference_to_task(consume_task, self._consume_tasks)

    async def _produce(self, name: int) -> None:
        self._logger.info(f"[Bot] Producer #{name} started")  # noqa: G004
        try:
            async for raw_message in self._signal.receive():
                self._logger.info(f"[Raw Message] {raw_message}")  # noqa: G004

                try:
                    message = await Message.parse(self._signal, raw_message)
                except UnknownMessageFormatError:
                    continue

                # Update groups if message is from an unknown group
                if (
                    message.is_group()
                    and self._groups_by_internal_id.get(message.group) is None
                ):
                    await self._detect_groups()

                await self._ask_commands_to_handle(message)

        except ReceiveMessagesError as e:
            # TODO: retry strategy  # noqa: TD002, TD003
            raise SignalBotError(f"Cannot receive messages: {e}")  # noqa: B904, EM102, TRY003

    def _should_react_for_contact(
        self,
        message: Message,
        contacts: list[str] | bool,  # noqa: FBT001
        group_ids: list[str] | bool,  # noqa: FBT001
    ) -> bool:
        """Is the command activated for a certain chat or group?"""
        # Case 1: Private message
        if message.is_private():
            # a) registered for all numbers
            if isinstance(contacts, bool) and contacts:
                return True

            # b) whitelisted numbers
            if isinstance(contacts, list) and message.source in contacts:
                return True

        # Case 2: Group message
        if message.is_group():
            # a) registered for all groups
            if isinstance(group_ids, bool) and group_ids:
                return True

            # b) whitelisted group ids
            group_id = self._groups_by_internal_id.get(message.group, {}).get("id")
            if isinstance(group_ids, list) and group_id and group_id in group_ids:
                return True

        return False

    def _should_react_for_lambda(
        self,
        message: Message,
        f: Callable[[Message], bool] | None = None,
    ) -> bool:
        if f is None:
            return True

        return f(message)

    async def _ask_commands_to_handle(self, message: Message) -> None:
        for command, contacts, group_ids, f in self.commands:
            if not self._should_react_for_contact(message, contacts, group_ids):
                continue

            if not self._should_react_for_lambda(message, f):
                continue

            await self._q.put((command, message, time.perf_counter()))

    async def _consume(self, name: int) -> None:
        self._logger.info(f"[Bot] Consumer #{name} started")  # noqa: G004
        while True:
            try:
                await self._consume_new_item(name)
            except Exception:  # noqa: BLE001, PERF203, S112
                continue

    async def _consume_new_item(self, name: int) -> None:
        command, message, t = await self._q.get()
        now = time.perf_counter()
        self._logger.info(
            f"[Bot] Consumer #{name} got new job in {now - t:0.5f} seconds"  # noqa: G004
        )

        # handle Command
        try:
            context = Context(self, message)
            await command.handle(context)
        except Exception:
            self._logger.exception(f"[{command.__class__.__name__}]")  # noqa: G004
            raise

        # done
        self._q.task_done()


class SignalBotError(Exception):
    pass



from __future__ import annotations

from pydantic import BaseModel


class LinkPreview(BaseModel):
    base64_thumbnail: str
    title: str
    description: str
    url: str

    #  This is the local filename for a received link preview
    id: str | None = None



from signalbot.api import ReceiveMessagesError, SendMessageError, SignalAPI
from signalbot.bot import SignalBot, enable_console_logging
from signalbot.command import Command, CommandError, regex_triggered, triggered
from signalbot.context import Context
from signalbot.message import Message, MessageType, Quote, UnknownMessageFormatError

__all__ = [
    "Command",
    "CommandError",
    "Context",
    "Message",
    "MessageType",
    "Quote",
    "ReceiveMessagesError",
    "SendMessageError",
    "SignalAPI",
    "SignalBot",
    "UnknownMessageFormatError",
    "enable_console_logging",
    "regex_triggered",
    "triggered",
]



from signalbot.utils.chat_testing import (
    ChatTestCase,
    DummyCommand,
    ReactMessageMock,
    ReceiveMessagesMock,
    SendMessagesMock,
    chat,
)

__all__ = [
    "ChatTestCase",
    "DummyCommand",
    "ReactMessageMock",
    "ReceiveMessagesMock",
    "SendMessagesMock",
    "chat",
]



try:  # noqa: SIM105
    import redis
except ModuleNotFoundError:
    pass

import json
import sqlite3
from abc import ABC, abstractmethod
from typing import Any


class Storage(ABC):
    @abstractmethod
    def exists(self, key: str) -> bool:
        pass

    @abstractmethod
    def read(self, key: str) -> Any:  # noqa: ANN401
        pass

    @abstractmethod
    def save(self, key: str, object: Any) -> None:  # noqa: A002, ANN401
        pass

    @abstractmethod
    def delete(self, key: str) -> None:
        pass


class StorageError(Exception):
    pass


class SQLiteStorage(Storage):
    def __init__(self, database: str = ":memory:", **kwargs):  # noqa: ANN003, ANN204
        self._sqlite = sqlite3.connect(database, **kwargs)
        self._sqlite.execute(
            "CREATE TABLE IF NOT EXISTS signalbot (key text unique, value text)",
        )

    def exists(self, key: str) -> bool:
        return self._sqlite.execute(
            "SELECT EXISTS(SELECT 1 FROM signalbot WHERE key = ?)",
            [key],
        ).fetchone()[0]

    def read(self, key: str) -> Any:  # noqa: ANN401
        try:
            result = self._sqlite.execute(
                "SELECT value FROM signalbot WHERE key = ?",
                [key],
            ).fetchone()[0]
            return json.loads(result)
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"SQLite load failed: {e}")  # noqa: B904, EM102, TRY003

    def save(self, key: str, object: Any) -> None:  # noqa: A002, ANN401
        try:
            value = json.dumps(object)
            self._sqlite.execute(
                "INSERT INTO signalbot VALUES (?, ?) ON CONFLICT(key) DO UPDATE SET value=?",  # noqa: E501
                [key, value, value],
            )
            self._sqlite.commit()
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"SQLite save failed: {e}")  # noqa: B904, EM102, TRY003

    def delete(self, key: str) -> None:
        try:
            self._sqlite.execute("DELETE FROM signalbot WHERE key = ?", [key])
            self._sqlite.commit()
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"SQLite delete failed: {e}")  # noqa: B904, EM102, TRY003


class RedisStorage(Storage):
    def __init__(self, host: str, port: int):  # noqa: ANN204
        self._redis = redis.Redis(host=host, port=port, db=0)

    def exists(self, key: str) -> bool:
        return self._redis.exists(key)

    def read(self, key: str) -> Any:  # noqa: ANN401
        try:
            result_bytes = self._redis.get(key)
            result_str = result_bytes.decode("utf-8")
            result_dict = json.loads(result_str)
            return result_dict  # noqa: RET504, TRY300
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"Redis load failed: {e}")  # noqa: B904, EM102, TRY003

    def save(self, key: str, object: Any) -> None:  # noqa: A002, ANN401
        try:
            object_str = json.dumps(object)
            self._redis.set(key, object_str)
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"Redis save failed: {e}")  # noqa: B904, EM102, TRY003

    def delete(self, key: str) -> None:
        try:
            self._redis.delete(key)
        except Exception as e:  # noqa: BLE001
            raise StorageError(f"Redis delete failed: {e}")  # noqa: B904, EM102, TRY003



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class Quote(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)  # Support fields in camel case

    id: int
    author: str
    author_number: str | None = None
    author_uuid: str
    text: str
    attachments: list[dict[str, Any]]



from __future__ import annotations

import json
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING

from signalbot.link_previews import LinkPreview
from signalbot.quote import Quote

if TYPE_CHECKING:
    from signalbot.api import SignalAPI


class MessageType(Enum):
    SYNC_MESSAGE = 1  # Message received in a linked device
    DATA_MESSAGE = 2  # Message received in a primary device
    EDIT_MESSAGE = 3  # Message received is an edit of a previous message
    DELETE_MESSAGE = 4  # Message received is a remote delete of a previous message
    READ_MESSAGE = 5  # User read some messages


@dataclass
class Message:
    # required
    source: str
    source_number: str | None
    source_uuid: str
    timestamp: int
    type: MessageType
    text: str
    # optional
    base64_attachments: list[str] = field(default_factory=list)
    attachments_local_filenames: list[str] = field(default_factory=list)
    view_once: bool = False
    link_previews: list[LinkPreview] = field(default_factory=list)
    group: str | None = None
    reaction: str | None = None
    mentions: list[str] = field(default_factory=list)
    quote: Quote | None = None
    read_messages: list[dict] | None = None
    target_sent_timestamp: int | None = None
    remote_delete_timestamp: int | None = None
    raw_message: str | None = None

    def recipient(self) -> str:
        # Case 1: Group chat
        if self.group:
            return self.group  # internal ID

        # Case 2: User chat
        return self.source

    def is_private(self) -> bool:
        return not bool(self.group)

    def is_group(self) -> bool:
        return bool(self.group)

    @classmethod
    def _extract_message_data(
        cls, envelope: dict
    ) -> tuple[MessageType, dict, int | None, int | None]:
        """Extract message type, data_message, and timestamps from envelope."""
        target_sent_timestamp = None

        if "syncMessage" in envelope:
            sync_message = envelope["syncMessage"]
            if sync_message == {}:
                raise UnknownMessageFormatError

            if "readMessages" in sync_message:
                message_type = MessageType.READ_MESSAGE
                data_message = {
                    "message": "",
                    "readMessages": sync_message["readMessages"],
                }
            else:
                message_type = MessageType.SYNC_MESSAGE
                data_message = sync_message["sentMessage"]

            if "editMessage" in data_message:
                message_type = MessageType.EDIT_MESSAGE
                target_sent_timestamp = data_message["editMessage"][
                    "targetSentTimestamp"
                ]
                data_message = data_message["editMessage"]["dataMessage"]

        elif "dataMessage" in envelope:
            message_type = MessageType.DATA_MESSAGE
            data_message = envelope["dataMessage"]

        elif "editMessage" in envelope:
            message_type = MessageType.EDIT_MESSAGE
            data_message = envelope["editMessage"]["dataMessage"]
            target_sent_timestamp = envelope["editMessage"]["targetSentTimestamp"]

        else:
            raise UnknownMessageFormatError

        remote_delete_timestamp = None
        if "remoteDelete" in data_message:
            message_type = MessageType.DELETE_MESSAGE
            remote_delete_timestamp = data_message["remoteDelete"]["timestamp"]

        return (
            message_type,
            data_message,
            target_sent_timestamp,
            remote_delete_timestamp,
        )

    @classmethod
    async def parse(cls, signal: SignalAPI, raw_message_str: str) -> Message:
        try:
            raw_message = json.loads(raw_message_str)
        except Exception as exc:
            raise UnknownMessageFormatError from exc

        envelope = raw_message["envelope"]
        # General attributes
        try:
            source = envelope["source"]
            source_uuid = envelope["sourceUuid"]
            timestamp = envelope["timestamp"]
        except Exception as exc:
            raise UnknownMessageFormatError from exc

        source_number = envelope.get("sourceNumber")

        message_type, data_message, target_sent_timestamp, remote_delete_timestamp = (
            cls._extract_message_data(envelope)
        )

        text = cls._parse_data_message(data_message)
        group = cls._parse_group_information(data_message)
        reaction = cls._parse_reaction(data_message)
        mentions = cls._parse_mentions(data_message)
        quote = cls._parse_quote(data_message)
        read_messages = data_message.get("readMessages")

        base64_attachments, attachments_local_filenames, link_previews = [], [], []
        view_once = False
        if signal.download_attachments:
            base64_attachments = await cls._parse_attachments(signal, data_message)
            attachments_local_filenames = cls._parse_attachments_local_filenames(
                data_message,
            )
            link_previews = await cls._parse_previews(signal, data_message)
            view_once = data_message.get("viewOnce", False)

        return cls(
            source,
            source_number,
            source_uuid,
            timestamp,
            message_type,
            text,
            base64_attachments=base64_attachments,
            attachments_local_filenames=attachments_local_filenames,
            view_once=view_once,
            link_previews=link_previews,
            group=group,
            reaction=reaction,
            mentions=mentions,
            quote=quote,
            read_messages=read_messages,
            target_sent_timestamp=target_sent_timestamp,
            remote_delete_timestamp=remote_delete_timestamp,
            raw_message=raw_message_str,
        )

    @classmethod
    async def _parse_attachments(cls, signal: SignalAPI, data_message: dict) -> str:
        if "attachments" not in data_message:
            return []

        return [
            await signal.get_attachment(attachment["id"])
            for attachment in data_message["attachments"]
        ]

    @classmethod
    def _parse_attachments_local_filenames(cls, data_message: dict) -> list[str]:
        if "attachments" not in data_message:
            return []

        # The ["id"] is the local filename and the ["filename"] is the remote filename
        return [attachment["id"] for attachment in data_message["attachments"]]

    @classmethod
    def _parse_data_message(cls, data_message: dict) -> str:
        try:
            return data_message["message"]
        except KeyError as exc:
            raise UnknownMessageFormatError from exc

    @classmethod
    def _parse_group_information(cls, message: dict) -> str:
        try:
            return message["groupInfo"]["groupId"]
        except KeyError:
            return None

    @classmethod
    def _parse_mentions(cls, data_message: dict) -> list:
        try:
            return data_message["mentions"]
        except KeyError:
            return []

    @classmethod
    def _parse_reaction(cls, message: dict) -> str:
        try:
            return message["reaction"]["emoji"]
        except KeyError:
            return None

    @classmethod
    def _parse_quote(cls, message: dict) -> Quote | None:
        try:
            return Quote.model_validate(message["quote"])
        except KeyError:
            return None

    def __str__(self) -> str:
        if self.text is None:
            return ""
        return self.text

    @classmethod
    async def _parse_previews(cls, signal: SignalAPI, data_message: dict) -> list:
        parsed_previews = []
        try:
            for preview in data_message["previews"]:
                base64_thumbnail = await signal.get_attachment(preview["image"]["id"])
                parsed_previews.append(
                    LinkPreview(
                        base64_thumbnail=base64_thumbnail,
                        title=preview["title"],
                        description=preview["description"],
                        url=preview["url"],
                        id=preview["image"]["id"],
                    ),
                )
        except KeyError:
            return []

        return parsed_previews


class UnknownMessageFormatError(Exception):
    pass



from __future__ import annotations

import functools
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, ParamSpec, TypeVar

T = TypeVar("T")
P = ParamSpec("P")

if TYPE_CHECKING:
    from signalbot.bot import SignalBot
    from signalbot.context import Context


def regex_triggered(
    *by: str | re.Pattern[str],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator_regex_triggered(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        async def wrapper_regex_triggered(
            *args: P.args, **kwargs: P.kwargs
        ) -> T | None:
            c: Context = args[1]
            text = c.message.text
            if not isinstance(text, str):
                return None
            matches = [bool(re.search(pattern, text)) for pattern in by]
            if True not in matches:
                return None
            return await func(*args, **kwargs)

        return wrapper_regex_triggered

    return decorator_regex_triggered


def triggered(
    *by: str, case_sensitive: bool = False
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator_triggered(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        async def wrapper_triggered(*args: P.args, **kwargs: P.kwargs) -> T | None:
            c: Context = args[1]
            text = c.message.text
            if not isinstance(text, str):
                return None

            by_words = by
            if not case_sensitive:
                text = text.lower()
                by_words = [t.lower() for t in by_words]
            if text not in by_words:
                return None

            return await func(*args, **kwargs)

        return wrapper_triggered

    return decorator_triggered


class Command(ABC):
    def __init__(self) -> None:
        # The bot attribute is assigned after calling bot.register(Command())
        self.bot: SignalBot | None = None

    def setup(self) -> None:
        return

    @abstractmethod
    async def handle(self, context: Context) -> None:
        pass


class CommandError(Exception):
    pass



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from signalbot.bot import SignalBot
    from signalbot.link_previews import LinkPreview
    from signalbot.message import Message


class Context:
    def __init__(self, bot: SignalBot, message: Message) -> None:
        self.bot = bot
        self.message = message

    async def send(  # noqa: PLR0913
        self,
        text: str,
        *,
        base64_attachments: list[str] | None = None,
        link_preview: LinkPreview | None = None,
        mentions: list[dict[str, Any]] | None = None,
        text_mode: str | None = None,
        view_once: bool = False,
    ) -> int:
        return await self.bot.send(
            self.message.recipient(),
            text,
            base64_attachments=base64_attachments,
            mentions=mentions,
            text_mode=text_mode,
            link_preview=link_preview,
            view_once=view_once,
        )

    async def edit(  # noqa: PLR0913
        self,
        text: str,
        edit_timestamp: int,
        *,
        base64_attachments: list[str] | None = None,
        link_preview: LinkPreview | None = None,
        mentions: list[dict[str, Any]] | None = None,
        text_mode: str | None = None,
        view_once: bool = False,
    ) -> int:
        return await self.bot.send(
            self.message.recipient(),
            text,
            base64_attachments=base64_attachments,
            mentions=mentions,
            text_mode=text_mode,
            edit_timestamp=edit_timestamp,
            link_preview=link_preview,
            view_once=view_once,
        )

    async def reply(  # noqa: PLR0913
        self,
        text: str,
        *,
        base64_attachments: list[str] | None = None,
        link_preview: LinkPreview | None = None,
        mentions: (
            list[dict[str, Any]] | None
        ) = None,  # [{ "author": "uuid" , "start": 0, "length": 1 }]
        text_mode: str | None = None,
        view_once: bool = False,
    ) -> int:
        send_mentions = self._convert_receive_mentions_into_send_mentions(
            self.message.mentions,
        )
        return await self.bot.send(
            self.message.recipient(),
            text,
            base64_attachments=base64_attachments,
            quote_author=self.message.source,
            quote_mentions=send_mentions,
            quote_message=self.message.text,
            quote_timestamp=self.message.timestamp,
            mentions=mentions,
            text_mode=text_mode,
            link_preview=link_preview,
            view_once=view_once,
        )

    async def react(self, emoji: str) -> None:
        await self.bot.react(self.message, emoji)

    async def receipt(self, receipt_type: Literal["read", "viewed"]) -> None:
        await self.bot.receipt(self.message, receipt_type)

    async def start_typing(self) -> None:
        await self.bot.start_typing(self.message.recipient())

    async def stop_typing(self) -> None:
        await self.bot.stop_typing(self.message.recipient())

    async def remote_delete(self, timestamp: int) -> int:
        return await self.bot.remote_delete(
            self.message.recipient(), timestamp=timestamp
        )

    def _convert_receive_mentions_into_send_mentions(
        self,
        mentions: list[dict[str, Any]] | None = None,
    ) -> list[dict[str, Any]] | None:
        if mentions is None:
            return None

        send_mentions = deepcopy(mentions)
        for mention in send_mentions:
            if "author" not in mention:
                mention["author"] = mention["uuid"]
        return send_mentions



import functools
import json
import time
import unittest
import uuid
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp

from signalbot.bot import Command, Context, SignalBot


def chat(*messages):  # noqa: ANN002, ANN201
    def decorator_chat(func):  # noqa: ANN001, ANN202
        signalbot_package = ".".join(__package__.split(".")[:-1])

        @functools.wraps(func)
        @patch(f"{signalbot_package}.SignalAPI.react", new_callable=ReactMessageMock)
        @patch(f"{signalbot_package}.SignalAPI.send", new_callable=SendMessagesMock)
        @patch(
            f"{signalbot_package}.SignalAPI.receive",
            new_callable=ReceiveMessagesMock,
        )
        async def wrapper_chat(*args, **kwargs):  # noqa: ANN002, ANN003, ANN202
            chat_test_case = args[0]
            receive_mock = args[1]

            receive_mock.define(messages)
            await chat_test_case.run_bot()

            value = func(*args, **kwargs)
            return value  # noqa: RET504

        return wrapper_chat

    return decorator_chat


class ChatTestCase(unittest.IsolatedAsyncioTestCase):
    signal_service = "127.0.0.1:8080"
    phone_number = "+49123456789"

    group_id = "group_id1="
    group_secret = "group.group_secret1="  # noqa: S105
    config = {  # noqa: RUF012
        "signal_service": signal_service,
        "phone_number": phone_number,
        "storage": {"type": "in-memory"},
    }

    def setUp(self):  # noqa: ANN201
        self.signal_bot = SignalBot(ChatTestCase.config)

    async def run_bot(self):  # noqa: ANN201
        PRODUCER_ID = 1337  # noqa: N806
        HANDLER_ID = 4444  # noqa: N806
        await self.signal_bot._produce(PRODUCER_ID)  # noqa: SLF001
        while self.signal_bot._q.qsize() > 0:  # noqa: SLF001
            await self.signal_bot._consume_new_item(HANDLER_ID)  # noqa: SLF001

    @classmethod
    def new_message(cls, text) -> str:  # noqa: ANN001
        timestamp = time.time()
        new_uuid = str(uuid.uuid4())
        message = {
            "envelope": {
                "source": ChatTestCase.phone_number,
                "sourceNumber": ChatTestCase.phone_number,
                "sourceUuid": new_uuid,
                "sourceName": "some_source_name",
                "sourceDevice": 1,
                "timestamp": timestamp,
                "syncMessage": {
                    "sentMessage": {
                        "timestamp": timestamp,
                        "message": text,
                        "expiresInSeconds": 0,
                        "viewOnce": False,
                        "mentions": [],
                        "attachments": [],
                        "contacts": [],
                        "groupInfo": {
                            "groupId": ChatTestCase.group_id,
                            "type": "DELIVER",
                        },
                        "destination": None,
                        "destinationNumber": None,
                        "destinationUuid": None,
                    },
                },
            },
        }
        return json.dumps(message)


class ReceiveMessagesMock(MagicMock):
    def define(self, messages: list):  # noqa: ANN201
        json_messages = [ChatTestCase.new_message(m) for m in messages]
        mock_iterator = AsyncMock()
        mock_iterator.__aiter__.return_value = json_messages
        self.return_value = mock_iterator


class SendMessagesMock(AsyncMock):
    def __init__(self, *args, **kwargs):  # noqa: ANN002, ANN003, ANN204
        super().__init__(*args, **kwargs)
        mock = AsyncMock()
        mock.return_value = {"timestamp": "1638715559464"}
        self.return_value = AsyncMock(
            spec=aiohttp.ClientResponse,
            status_code=201,
            json=mock,
        )

    def results(self) -> list:
        return self._extract_responses()

    def _extract_responses(self):  # noqa: ANN202
        results = []
        for args in self.call_args_list:
            results.append(args[0])  # noqa: PERF401
        return results


class ReactMessageMock(AsyncMock):
    def results(self) -> list:
        return self._extract_responses()

    def _extract_responses(self):  # noqa: ANN202
        results = []
        for args in self.call_args_list:
            results.append(args[0])  # noqa: PERF401
        return results


class DummyCommand(Command):
    async def handle(self, context: Context):  # noqa: ANN201
        pass



from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Any, Literal

import aiohttp
import websockets

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


class SignalAPI:
    def __init__(  # noqa: ANN204
        self,
        signal_service: str,
        phone_number: str,
        download_attachments: bool = True,  # noqa: FBT001, FBT002
    ):
        self.phone_number = phone_number
        self._signal_api_uris = SignalAPIURIs(
            signal_service=signal_service,
            phone_number=phone_number,
        )
        self.download_attachments = download_attachments

    async def receive(self) -> AsyncIterator[str]:
        try:
            uri = self._signal_api_uris.receive_ws_uri()
            self.connection = websockets.connect(uri, ping_interval=None)
            async with self.connection as websocket:
                async for raw_message in websocket:
                    yield raw_message

        except Exception as e:  # noqa: BLE001
            raise ReceiveMessagesError(e)  # noqa: B904

    async def send(  # noqa: C901, PLR0913
        self,
        receiver: str,
        message: str,
        *,
        base64_attachments: list | None = None,
        link_preview: dict[str, Any] | None = None,
        quote_author: str | None = None,
        quote_mentions: list | None = None,
        quote_message: str | None = None,
        quote_timestamp: int | None = None,
        mentions: list[dict[str, Any]] | None = None,
        text_mode: str | None = None,
        edit_timestamp: int | None = None,
        view_once: bool = False,
    ) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.send_rest_uri()
        if base64_attachments is None:
            base64_attachments = []

        payload = {
            "base64_attachments": base64_attachments,
            "message": message,
            "number": self.phone_number,
            "recipients": [receiver],
        }

        if quote_author:
            payload["quote_author"] = quote_author
        if quote_mentions:
            payload["quote_mentions"] = quote_mentions
        if quote_message:
            payload["quote_message"] = quote_message
        if quote_timestamp:
            payload["quote_timestamp"] = quote_timestamp
        if mentions:
            payload["mentions"] = mentions
        if text_mode:
            payload["text_mode"] = text_mode
        if edit_timestamp:
            payload["edit_timestamp"] = edit_timestamp
        if link_preview:
            payload["link_preview"] = link_preview
        if view_once:
            payload["view_once"] = True

        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.post(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
            KeyError,
        ):
            raise SendMessageError  # noqa: B904

    async def react(
        self,
        recipient: str,
        reaction: str,
        target_author: str,
        timestamp: int,
    ) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.react_rest_uri()
        payload = {
            "recipient": recipient,
            "reaction": reaction,
            "target_author": target_author,
            "timestamp": timestamp,
        }
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.post(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise ReactionError  # noqa: B904

    async def receipt(
        self,
        recipient: str,
        receipt_type: Literal["read", "viewed"],
        timestamp: int,
    ) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.receipts_rest_uri()
        payload = {
            "recipient": recipient,
            "receipt_type": receipt_type,
            "timestamp": timestamp,
        }
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.post(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise ReactionError  # noqa: B904

    async def start_typing(self, receiver: str) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.typing_indicator_uri()
        payload = {
            "recipient": receiver,
        }
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.put(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise StartTypingError  # noqa: B904

    async def stop_typing(self, receiver: str) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.typing_indicator_uri()
        payload = {
            "recipient": receiver,
        }
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.delete(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise StopTypingError  # noqa: B904

    async def get_groups(self) -> list[dict[str, Any]]:
        uri = self._signal_api_uris.groups_uri()
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.get(uri)
                resp.raise_for_status()
                return await resp.json()
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise GroupsError  # noqa: B904

    async def get_attachment(self, attachment_id: str) -> str:
        uri = f"{self._signal_api_uris.attachment_rest_uri()}/{attachment_id}"
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.get(uri)
                resp.raise_for_status()
                content = await resp.content.read()
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise GetAttachmentError  # noqa: B904

        base64_bytes = base64.b64encode(content)
        base64_string = str(base64_bytes, encoding="utf-8")

        return base64_string  # noqa: RET504

    async def delete_attachment(self, attachment_id: str) -> str:
        uri = f"{self._signal_api_uris.attachment_rest_uri()}/{attachment_id}"
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.delete(uri)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise GetAttachmentError  # noqa: B904

    async def update_contact(
        self,
        receiver: str,
        expiration_in_seconds: int | None = None,
        name: str | None = None,
    ) -> None:
        uri = self._signal_api_uris.contacts_uri()
        payload = {"recipient": receiver}

        if expiration_in_seconds is not None:
            payload["expiration_in_seconds"] = expiration_in_seconds

        if name is not None:
            payload["name"] = name

        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.put(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise ContactUpdateError  # noqa: B904

    async def update_group(
        self,
        group_id: str,
        base64_avatar: str | None = None,
        description: str | None = None,
        expiration_in_seconds: int | None = None,
        name: str | None = None,
    ) -> None:
        uri = self._signal_api_uris.group_id_uri(group_id)
        payload = {}

        if base64_avatar is not None:
            payload["base64_avatar"] = base64_avatar

        if description is not None:
            payload["description"] = description

        if expiration_in_seconds is not None:
            payload["expiration_time"] = expiration_in_seconds

        if name is not None:
            payload["name"] = name

        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.put(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise ContactUpdateError  # noqa: B904

    async def health_check(self) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.health_check_uri()
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.get(uri)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise HealthCheckError  # noqa: B904

    async def check_signal_service(self) -> bool:
        self._signal_api_uris.use_https = True
        try:
            return (await self.health_check()).status == 204  # noqa: PLR2004
        except HealthCheckError:
            self._signal_api_uris.use_https = False
            try:
                return (await self.health_check()).status == 204  # noqa: PLR2004
            except HealthCheckError:
                return False

    async def get_signal_cli_rest_api_version(self) -> str:
        uri = self._signal_api_uris.about_rest_uri()
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.get(uri)
                resp.raise_for_status()
                return (await resp.json())["version"]
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise AboutError  # noqa: B904

    async def remote_delete(
        self, receiver: str, timestamp: int
    ) -> aiohttp.ClientResponse:
        uri = self._signal_api_uris.remote_delete_uri()
        payload = {
            "recipient": receiver,
            "timestamp": timestamp,
        }
        try:
            async with aiohttp.ClientSession() as session:
                resp = await session.delete(uri, json=payload)
                resp.raise_for_status()
                return resp
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ):
            raise RemoteDeleteError  # noqa: B904


class SignalAPIURIs:
    def __init__(self, signal_service: str, phone_number: str, use_https: bool = True):  # noqa: ANN204, FBT001, FBT002
        self.signal_service = signal_service
        self.phone_number = phone_number
        self.use_https = use_https

    @property
    def https_or_http(self) -> str:
        return "https" if self.use_https else "http"

    @property
    def wss_or_ws(self) -> str:
        return "wss" if self.use_https else "ws"

    def attachment_rest_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/attachments"

    def receive_ws_uri(self) -> str:
        return (
            f"{self.wss_or_ws}://{self.signal_service}/v1/receive/{self.phone_number}"
        )

    def send_rest_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v2/send"

    def react_rest_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/reactions/{self.phone_number}"

    def typing_indicator_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/typing-indicator/{self.phone_number}"

    def groups_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/groups/{self.phone_number}"

    def group_id_uri(self, group_id: str) -> str:
        return self.groups_uri() + "/" + group_id

    def contacts_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/contacts/{self.phone_number}"

    def health_check_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/health"

    def receipts_rest_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/receipts/{self.phone_number}"

    def remote_delete_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/remote-delete/{self.phone_number}"

    def about_rest_uri(self) -> str:
        return f"{self.https_or_http}://{self.signal_service}/v1/about"


class ReceiveMessagesError(Exception):
    pass


class SendMessageError(Exception):
    pass


class TypingError(Exception):
    pass


class StartTypingError(TypingError):
    pass


class StopTypingError(TypingError):
    pass


class ReactionError(Exception):
    pass


class GroupsError(Exception):
    pass


class GetAttachmentError(Exception):
    pass


class ContactUpdateError(Exception):
    pass


class HealthCheckError(Exception):
    pass


class RemoteDeleteError(Exception):
    pass


class SignalServiceConnectionError(Exception):
    pass


class AboutError(Exception):
    pass
